//
// sharpflash
//
// This program is used to read and write NAND flash
// on the Reciva Radio Module, via a JTAG Wiggler 
// interface.
//
// It is based on sjf2410, and uses files from
//   sjf2410 - 
//
// $Id: sharpflash.c 223 2007-10-17 21:46:52Z  $
//
// This program is very slow, as it uses the boundary
// scan path to directly drive the nand flash.  What it
// needs to do really, is use mem-read and mem-write, and
// make use of the ARM NAND Flash Controller.
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "def.h"
#include "jtag.h"
#include "ppt.h"
#include "pin2410.h"
#include "k9f2808.h"

void OpenPpt(void);

int main(int argc,char *argv[])
{
	U32 id32 ;
	U16 dev ;
	U32 start, len ;

	if (argc!=5 && argc!=2) {
		printf("\nsharpflash r|w|b file.bin [ start len ]\n\n") ;
		printf("  r|w|b    - read, write or check bad NAND Flash data\n") ;
		printf("  file.bin - destination / source filename\n") ;
		printf("  start    - hex start position in NAND for read/write\n") ;
		printf("             start must be a multiple of 4000(hex)\n") ;
		printf("  len      - hex length to read/write. if file is too\n") ;
		printf("             short, NAND will be filled with 0xFF\n") ; 
		printf("             len must be a multiple of 4000(hex)\n\n") ;
		printf(" Output data contains:\n\n") ;
		printf("   w - page written OK\n") ;
		printf("   r - page read OK\n") ;
		printf("   . - page check OK\n") ;
		printf("   b - page/block identified as bad\n") ;
		printf("   B - page/block has just been marked bad\n\n") ;
		printf("\nhttp://www.sharpfin.zevv.nl/\n$Revision: 223 $\n\n") ;
		return 0;
	}

	printf("\nSharpfin Flash Programmer - http://www.sharpfin.zevv.nl/\n\n") ;

	// Open Parallel Port
	validPpt = GetValidPpt();
	if(!validPpt) {
		printf("Unable to find a parallel port\n");
		return 0 ;
	}
	SetPptCompMode();

	// Connect to the Processor
	id32=JTAG_ReadId();
	if (id32!=0x0032409d) {
		printf("Unable to find S3C2410 processor on JTAG Cable, found %X\n", id32) ;
		return 0 ;
	} else {
		printf("Found S3C2410 processor on JTAG Cable\n") ;
	}

	S2410_InitCell() ;

	K9F2808_JtagInit();
	K9F2808_init();

	// Check the NAND Flash
	dev=K9F2808_checkid();
	if (dev!=0xEC75) {
		printf("Unable to find K9F2808UOC on processor, found 0x%04X\n", dev) ;
		return 0 ;
	} else {
		printf("Found K9F2808UOC on processor\n") ;
	}


	if (argv[1][0]=='r' || argv[1][0]=='w') {
	
		if (argc!=5) {
			printf("sharpflash: start and/or length not supplied\n") ;
			return 0 ;
		}

		sscanf(argv[3], "%X", &start) ;
		sscanf(argv[4], "%X", &len) ;

		// Check the passed parameters
	
		if ( ((start/0x4000)*0x4000) != start ) {
			printf("start address must be a multiple of 4000(hex)\n") ;
			return 0 ;
		}

		if ( ((len/0x4000)*0x4000) != len ) {
			printf("len must be a multiple of 4000(hex)\n") ;
			return 0 ;
		}

		if (argv[1][0]=='r') {
	
			// Read data
			K9F2808_read(argv[2], start, len) ;

		} else if (argv[1][0]=='w') {

			// Write data
			K9F2808_write(argv[2], start, len) ;
		}

	} else if (argv[1][0]=='b') {

		// Check Bad Blocks
		K9F2808_badcheck() ;


	} else {

		printf("Unknown command: %s\n", argv[1]) ;

	}

	printf("\n") ;
	return 0 ;

}

